"""This module contains code that handles comparing modules"""

import ast
import logging
from typing import List, Optional, Dict

import pyff.classes as pc
import pyff.functions as pf
import pyff.imports as pi
from pyff.kitchensink import hl


LOGGER = logging.getLogger(__name__)


class ModuleSummary:  # pylint: disable=too-few-public-methods
    """Holds summary information about a module"""

    def __init__(self, name: str, node: ast.Module) -> None:
        self.name: str = name
        self.node: ast.Module = node


class ModulePyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between two Python modules"""

    def __init__(
        self,
        imports: Optional[pi.ImportsPyfference] = None,
        classes: Optional[pc.ClassesPyfference] = None,
        functions: Optional[pf.FunctionsPyfference] = None,
    ) -> None:

        self.other: List = []
        self.imports: Optional[pi.ImportsPyfference] = imports
        self.classes: Optional[pc.ClassesPyfference] = classes
        self.functions: Optional[pf.FunctionsPyfference] = functions

    def __str__(self):
        changes = [self.imports, self.classes, self.functions] + self.other
        return "\n".join([str(change) for change in changes if change is not None])

    def simplify(self) -> Optional["ModulePyfference"]:
        """Cleans empty differences, empty sets etc. after manipulation"""
        if self.imports is not None:
            self.imports = self.imports.simplify()

        if self.classes is not None:
            self.classes = self.classes.simplify()

        if self.functions is not None:
            self.functions = self.functions.simplify()

        return self if (self.functions or self.classes or self.imports or self.other) else None


class ModulesPyfference:  # pylint: disable=too-few-public-methods
    """Holds difference between modules in a package"""

    def __init__(
        self,
        removed: Dict[str, ModuleSummary],
        changed: Dict[str, ModulePyfference],
        new: Dict[str, ModuleSummary],
    ) -> None:
        self.removed: Dict[str, ModuleSummary] = removed
        self.changed: Dict[str, ModulePyfference] = changed
        self.new: Dict[str, ModuleSummary] = new

    def __str__(self):
        return "\n".join(
            [
                f"Module {hl(module)} changed:\n  " + str(change).replace("\n", "\n  ")
                for module, change in self.changed.items()
            ]
        )

    def __repr__(self):
        return (
            f"ModulesPyfference(removed={repr(self.removed)}, "
            f"changed={repr(self.changed)}, new={repr(self.new)})"
        )

    def __bool__(self):
        return bool(self.removed or self.changed or self.new)


def pyff_module(old: ast.Module, new: ast.Module) -> Optional[ModulePyfference]:
    """Return difference between two Python modules, or None if they are identical"""
    old_imports = pi.ImportedNames.extract(old)
    new_imports = pi.ImportedNames.extract(new)
    imports = pi.pyff_imports(old, new)
    classes = pc.pyff_classes(old, new, old_imports, new_imports)
    functions = pf.pyff_functions(old, new, old_imports, new_imports)

    if imports or classes or functions:
        LOGGER.debug("Modules differ")
        pyfference = ModulePyfference(imports, classes, functions)
        return pyfference

    LOGGER.debug("Modules are identical")
    return None


def pyff_module_code(old: str, new: str) -> Optional[ModulePyfference]:
    """Return difference between two Python modules, or None if they are identical"""
    # pylint: disable=unused-variable
    old_ast = ast.parse(old)
    new_ast = ast.parse(new)
    return pyff_module(old_ast, new_ast)



"""This module contains code that handles comparing packages"""
import logging
import pathlib
import ast

from typing import Optional, Iterable, FrozenSet, Set
from astroid.modutils import get_module_files

import pyff.modules as pm

LOGGER = logging.getLogger(__name__)


class PackagePyfference:  # pylint: disable=too-few-public-methods
    """Holds differences between two Python packages"""

    def __init__(self, modules: Optional[pm.ModulesPyfference]) -> None:
        self.modules: Optional[pm.ModulesPyfference] = modules

    def __str__(self):
        return str(self.modules)

    def __repr__(self):
        return f"PackagePyfference(modules={repr(self.modules)})"


def extract_modules(files: Iterable[pathlib.Path], package_path: pathlib.Path) -> FrozenSet[str]:
    """Extract direct modules of a packages (i.e. not modules of subpackages"""
    LOGGER.debug(str(files))
    LOGGER.debug(str(package_path))
    return frozenset({module.name for module in files if module.parents[0] == package_path})


def _compare_module_in_packages(
    module: pathlib.Path, old_package: pathlib.Path, new_package: pathlib.Path
) -> Optional[pm.ModulePyfference]:
    """Compare one module in two packages"""
    old_module = old_package / module
    new_module = new_package / module

    return pm.pyff_module_code(old_module.read_text(), new_module.read_text())


def _summarize_module_in_package(module: pathlib.Path, package: pathlib.Path) -> pm.ModuleSummary:
    full_path = package / module
    module_ast = ast.parse(full_path.read_text())
    return pm.ModuleSummary(str(module), module_ast)


def pyff_package(
    old_package: pathlib.Path, new_package: pathlib.Path
) -> Optional[PackagePyfference]:
    """Given *paths* to two versions of a package, return differences between them"""
    old_files: Set[pathlib.Path] = {
        pathlib.Path(module) for module in get_module_files(old_package, ())
    }
    new_files: Set[pathlib.Path] = {
        pathlib.Path(module) for module in get_module_files(new_package, ())
    }

    LOGGER.debug("Files of the old package %s: %s", old_package, old_files)
    LOGGER.debug("Files of the new package %s: %s", new_package, new_files)

    old_modules: Set[pathlib.Path] = {
        pathlib.Path(module) for module in extract_modules(old_files, old_package)
    }
    new_modules: Set[pathlib.Path] = {
        pathlib.Path(module) for module in extract_modules(new_files, new_package)
    }

    LOGGER.debug("Old modules: %s", str(old_modules))
    LOGGER.debug("New modules: %s", str(new_modules))

    removed = old_modules - new_modules
    new = new_modules - old_modules
    both = old_modules.intersection(new_modules)

    LOGGER.debug("Removed modules: %s", str(removed))
    LOGGER.debug("Modules in both packages: %s", str(both))
    LOGGER.debug("New modules: %s", str(new))

    removed_summaries = {
        str(module): _summarize_module_in_package(module, old_package) for module in removed
    }
    new_summaries = {
        str(module): _summarize_module_in_package(module, new_package) for module in new
    }
    changed = {
        str(module): change
        for module, change in [
            (module, _compare_module_in_packages(module, old_package, new_package))
            for module in both
        ]
        if change is not None
    }
    modules = pm.ModulesPyfference(removed_summaries, changed, new_summaries)

    if modules:
        return PackagePyfference(modules)

    return None



"""This module contains code that handles comparing imports"""

import collections.abc
import types
from typing import Set, Dict, Union, Optional, FrozenSet, Mapping, cast
import ast
import logging
from pyff.kitchensink import hl, hlistify, pluralize

ImportNode = Union[ast.Import, ast.ImportFrom]  # pylint: disable=invalid-name

LOGGER = logging.getLogger(__name__)


class ImportedName:
    """Represents a single imported name"""

    def __init__(self, name: str, node: ImportNode, alias: ast.alias) -> None:
        self.name: str = name
        self.node: ImportNode = node
        self.alias: ast.alias = alias

    def __repr__(self):  # pragma: no cover
        return f"ImportedName(name={self.name} node={self.node} alias={self.alias}"

    def is_import(self) -> bool:
        """Returns True if name was imported with `import X` statement"""
        return isinstance(self.node, ast.Import)

    def is_import_from(self) -> bool:
        """Returns True if name was imported with `from Y import X` statement"""
        return isinstance(self.node, ast.ImportFrom)

    @property
    def canonical_name(self) -> str:
        """Returns whole name.

        Example: For 'join' imported by 'from os.path import join', returns 'os.path.join'"""
        if isinstance(self.node, ast.Import):
            return self.alias.name
        elif isinstance(self.node, ast.ImportFrom):
            return f"{self.node.module}.{self.alias.name}"

        raise Exception("Node should always be one of {Import, ImportFrom}")  # pragma: no cover

    @property
    def canonical_ast(self) -> Union[ast.Name, ast.Attribute]:
        """Returns AST node for the full name

        Example: For 'join' imported by 'from os.path import join', returns AST of 'os.path.join'"""
        node: Union[ast.Name, ast.Attribute]

        if isinstance(self.node, ast.Import):
            items = self.alias.name.split(".")
            node = ast.Name(id=items.pop(0), ctx=ast.Load())
            while items:
                node = ast.Attribute(value=node, attr=items.pop(0), ctx=ast.Load())
            return node
        elif isinstance(self.node, ast.ImportFrom):
            if self.node.module is None:
                raise Exception(
                    "ast.ImportFrom has module attribute set to None"
                )  # pragma: no cover
            items = self.node.module.split(".") + [self.alias.name]
            node = ast.Name(id=items.pop(0), ctx=ast.Load())
            while items:
                node = ast.Attribute(value=node, attr=items.pop(0), ctx=ast.Load())
            return node

        raise Exception("Node should always be one of {Import, ImportFrom}")  # pragma: no cover

    def __str__(self):
        return self.name


class FromImportPyfference:
    """Represents difference in `from X import Y` between two ImportedNames"""

    def __init__(self):
        self._new: Dict[str, Set[ImportedName]] = {}
        self._removed: Dict[str, Set[ImportedName]] = {}
        self._new_modules: Set[str] = set()
        self._removed_modules: Set[str] = set()

    @property
    def new(self) -> Mapping[str, Set[ImportedName]]:
        """Returns a read-only mapping of new imported-from names"""
        return types.MappingProxyType(self._new)

    @property
    def removed(self) -> Mapping[str, Set[ImportedName]]:
        """Returns a read-only mapping of removed imported-from names"""
        return types.MappingProxyType(self._removed)

    @property
    def new_modules(self) -> FrozenSet[str]:
        """Returns a read-only set of new modules imported via `from X import Y` statements"""
        return frozenset(self._new_modules)

    @property
    def removed_modules(self) -> FrozenSet[str]:
        """Returns a read-only set of removed modules imported via `from X import Y` statements"""
        return frozenset(self._removed_modules)

    def add_new(self, node: ImportedName) -> None:
        """Add new name imported by `from X import y` statement"""
        if not node.is_import_from():
            raise ValueError(
                "FromImportPyfference can only handle ImportFrom nodes"
            )  # pragma: no cover

        module = cast(ast.ImportFrom, node.node).module

        if module is None:
            raise Exception("ast.ImportFrom has `module` attribute set to None")  # pragma: no cover

        if module not in self._new:
            self._new[module] = set()
        self._new[module].add(node)

    def add_removed(self, node: ImportedName) -> None:
        """Add removed name imported by `from X import y` statement"""
        if not node.is_import_from():
            raise ValueError(
                "FromImportPyfference can only handle ImportFrom nodes"
            )  # pragma: no cover

        module = cast(ast.ImportFrom, node.node).module
        if module is None:
            raise Exception("ast.ImportFrom has `module` attribute set to None")  # pragma: no cover

        if module not in self._removed:
            self._removed[module] = set()
        self._removed[module].add(node)

    def add_new_modules(self, modules: Set[str]) -> None:
        """Add new modules imported via `from X import Y` statements"""
        self._new_modules.update(modules)

    def add_removed_modules(self, modules: Set[str]) -> None:
        """Add removed modules imported via `from X import Y` statements"""
        self._removed_modules.update(modules)

    def delete_new_module(self, module: str) -> None:
        """Delete new module imported via `from X import Y` statements"""
        self._new_modules.discard(module)
        if module in self._new:
            del self._new[module]

    def delete_removed_module(self, module: str) -> None:
        """Delete removed module imported via `from X import Y` statements"""
        self._removed_modules.discard(module)
        if module in self._removed:
            del self._removed[module]

    def __bool__(self):
        return bool(self._new or self.removed or self.new_modules or self.removed_modules)


class ImportsPyfference:
    """Represent difference between two ImportedNames."""

    def __init__(self):
        self._new_imports: Set[ImportedName] = set()
        self._removed_imports: Set[ImportedName] = set()
        self.fromimports: FromImportPyfference = FromImportPyfference()
        self._changed_to_fromimport: Dict[str, Set[ImportedName]] = {}
        self._changed_to_import: Dict[str, Set[ImportedName]] = {}

    def __bool__(self):
        return bool(
            (
                self._new_imports
                or self._removed_imports
                or self.fromimports
                or self._changed_to_fromimport
                or self._changed_to_import
            )
        )

    @property
    def new_imports(self) -> FrozenSet[ImportedName]:
        """Returns a read-only set of new imported names"""
        return frozenset(self._new_imports)

    @property
    def removed_imports(self) -> FrozenSet[ImportedName]:
        """Returns a read-only set of removed imported names"""
        return frozenset(self._removed_imports)

    def simplify(self) -> Optional["ImportsPyfference"]:
        """Cleans empty differences, empty sets etc. after manipulation"""
        return self if self else None

    def new_import(self, node: ImportedName) -> None:
        """Add a new imported name"""
        self._new_imports.add(node)

    def removed_import(self, node: ImportedName) -> None:
        """Add a removed imported name"""
        self._removed_imports.add(node)

    def new_from_import(self, node: ImportedName) -> None:
        """Add a new name imported via `from X import Y` statement"""
        if not node.is_import_from():
            raise ValueError(
                "ImportsPyfference.new_from_import can only handle ImportFrom nodes"
            )  # pragma: no cover

        if cast(ast.ImportFrom, node.node).module:
            self.fromimports.add_new(node)

    def removed_from_import(self, node: ImportedName) -> None:
        """Add a removed name imported via `from X import Y` statement"""
        if not node.is_import_from():
            raise ValueError(
                "ImportsPyfference.new_from_import can only handle ImportFrom nodes"
            )  # pragma: no cover

        if cast(ast.ImportFrom, node.node).module:
            self.fromimports.add_removed(node)

    def new_fromimport_modules(self, modules: Set[str]) -> None:
        """Add new modules imported via `from X import Y` statement"""
        self.fromimports.add_new_modules(modules)

    def removed_fromimport_modules(self, modules: Set[str]) -> None:
        """Add removed modules imported via `from X import Y` statement"""
        self.fromimports.add_removed_modules(modules)

    def reduce(self) -> None:
        """Find special cases and other reductions in the differences

        (1) Find matching names imported by different import statements and
            create special records for these changes.
            Example: `from os import path` in one version and `import os` in another"""
        for name in set(self._new_imports):
            if name.name in self.fromimports.removed_modules:
                LOGGER.debug(
                    f"New module has 'import {name}' "
                    f"and old module had 'from {name} import ...': "
                    f"Adding a change record"
                )
                self._new_imports.discard(name)
                self._changed_to_import[name.name] = self.fromimports.removed[name.name]
                self.fromimports.delete_removed_module(name.name)

        for name in set(self._removed_imports):
            if name.name in self.fromimports.new_modules:
                self._removed_imports.discard(name)
                self._changed_to_fromimport[name.name] = self.fromimports.new[name.name]
                self.fromimports.delete_new_module(name.name)
                LOGGER.debug(
                    f"Old module had 'import {name}' and "
                    f"new module has 'from {name} import ...': "
                    f"Adding a change record"
                )

    def __str__(self):
        lines = []
        removed_imports = sorted([name.name for name in self.removed_imports])
        if removed_imports:
            packages = pluralize("package", removed_imports)
            names = hlistify(removed_imports)
            lines.append(f"Removed import of {packages} {names}")

        new_imports = sorted([name.name for name in self.new_imports])
        if new_imports:
            packages = pluralize("package", new_imports)
            names = hlistify(new_imports)
            lines.append(f"New imported {packages} {names}")

        for module, names in self.fromimports.removed.items():
            removed_names = sorted([str(name) for name in names])
            hl_removed_names = hlistify(removed_names)
            if module in self.fromimports.removed_modules:
                lines.append(f"Removed import of {hl_removed_names} from removed {hl(module)}")
            else:
                lines.append(f"Removed import of {hl_removed_names} from {hl(module)}")

        for module, names in self.fromimports.new.items():
            new_names = sorted([str(name) for name in names])
            if module in self.fromimports.new_modules:
                lines.append(f"New imported {hlistify(new_names)} from new {hl(module)}")
            else:
                lines.append(f"New imported {hlistify(new_names)} from {hl(module)}")

        for module, names in self._changed_to_fromimport.items():
            new_names = sorted([str(name) for name in names])
            lines.append(
                f"New imported {hlistify(new_names)} from {hl(module)} "
                f"(previously, full {hl(module)} was imported)"
            )

        for module, names in self._changed_to_import.items():
            new_names = sorted([str(name) for name in names])
            was = "was" if len(new_names) == 1 else "were"
            lines.append(
                f"New imported package {hl(module)} "
                f"(previously, only {hlistify(new_names)} "
                f"{was} imported from {hl(module)})"
            )

        return "\n".join(lines)


class ImportedNames(collections.abc.Mapping):  # pylint: disable=too-few-public-methods
    """Dictionary mapping external names to appropriate ImportedName"""

    @staticmethod
    def extract(code: ast.Module) -> "ImportedNames":
        """Extracts ImportedNames from a Module"""
        import_walker = ImportExtractor()
        import_walker.visit(code)
        return import_walker.names

    @staticmethod
    def compare(old: "ImportedNames", new: "ImportedNames") -> Optional[ImportsPyfference]:
        """Compare two sets of imported names."""
        LOGGER.debug("Comparing ImportedNames")
        change = ImportsPyfference()
        for name, node in new.names.items():
            if name not in old.names:
                LOGGER.debug(f"New name '{name}' not present in old names")
                if node.is_import():
                    change.new_import(node)
                elif node.is_import_from():
                    change.new_from_import(node)

        for name, node in old.names.items():
            if name not in new.names:
                LOGGER.debug(f"Old name '{name}' not present in new names")
                if node.is_import():
                    change.removed_import(node)
                elif node.is_import_from():
                    change.removed_from_import(node)

        change.new_fromimport_modules(new.from_modules - old.from_modules)
        LOGGER.debug(f"New modules from which names were imported: " f"{change.fromimports.new}")
        change.removed_fromimport_modules(old.from_modules - new.from_modules)
        LOGGER.debug(
            f"Removed modules from which names were imported: " f"{change.fromimports.removed}"
        )

        change.reduce()

        return change if change else None

    def __init__(self) -> None:
        self.names: Dict[str, ImportedName] = {}
        self.from_modules: Set[str] = set()

    def __getitem__(self, item):
        return self.names[item]

    def __iter__(self):
        yield from self.names

    def __len__(self):
        return len(self.names)

    def __repr__(self):  # pragma: no cover
        return f"ImportedNames(names={self.names}, from_modules={self.from_modules}"

    def _add(self, name: ast.alias, node: ImportNode) -> None:
        if name.asname is not None:
            self.names[name.asname] = ImportedName(name.asname, node, alias=name)
        else:
            self.names[name.name] = ImportedName(name.name, node, alias=name)

    def add_import(self, node: ast.Import) -> None:
        """Add a 'import X, Y' statement"""
        for name in node.names:
            self._add(name, node)

    def add_importfrom(self, node: ast.ImportFrom) -> None:
        """Add a 'from X import Y' statement"""
        for name in node.names:
            self._add(name, node)
        if node.module:
            self.from_modules.add(node.module)


class ImportExtractor(ast.NodeVisitor):
    """Extracts information about import and 'import from' statements"""

    def __init__(self) -> None:
        self.names = ImportedNames()
        super(ImportExtractor, self).__init__()

    def visit_Import(self, node):  # pylint: disable=invalid-name
        """Save information about `import X, Y` statements"""
        self.names.add_import(node)

    def visit_ImportFrom(self, node):  # pylint: disable=invalid-name
        """Save information about `from x import y` statements"""
        self.names.add_importfrom(node)


def pyff_imports(old: ast.Module, new: ast.Module) -> Optional[ImportsPyfference]:
    """Return differences in import statements in two modules"""
    old_walker = ImportExtractor()
    new_walker = ImportExtractor()

    old_walker.visit(old)
    new_walker.visit(new)

    difference = ImportedNames.compare(old_walker.names, new_walker.names)

    return difference if difference else None


def pyff_imports_code(old_code: str, new_code: str) -> Optional[ImportsPyfference]:
    """Return differences in import statements in two modules"""
    old_ast = ast.parse(old_code)
    new_ast = ast.parse(new_code)

    return pyff_imports(old_ast, new_ast)



"""This module contains code that handles comparing statements"""

import ast
import logging
import copy
from typing import Set, Optional, Dict
import pyff.imports as pi
from pyff.kitchensink import hl


LOGGER = logging.getLogger(__name__)


class SingleExternalNameUsageChange:
    """Represents a single external name usage change in a statement."""

    # pylint: disable=too-few-public-methods

    def __init__(self, old: str, new: str) -> None:
        self.old: str = old
        self.new: str = new

        self._key = (old, new)

    def __eq__(self, other):
        return (
            isinstance(other, SingleExternalNameUsageChange)
            and self.old == other.old
            and self.new == other.new
        )

    def __hash__(self):
        return hash(self._key)

    def __repr__(self):  # pragma: no cover
        return f"SingleExternalNameUsageChange(old={self.old}, new={self.new})"

    def __str__(self):
        return f"References of {hl(self.old)} were changed to {hl(self.new)}"


class ExternalNameUsageChange:
    """Represents case where two statements differ only in how they use external names.

    Example:
        (1) something = os.path.join(...)
        (2) something = path.join(...)

        Provided statement (1) runs in context where 'import os.path' was
        executed and (2) runs in context where 'from os import path' was imported,
        the two statements are semantically identical."""

    # pylint: disable=too-few-public-methods

    def __init__(self, changes: Set[SingleExternalNameUsageChange]) -> None:
        self.changes: Set[SingleExternalNameUsageChange] = changes

    def __repr__(self):  # pragma: no cover
        return f"ExternalNameUsageChange(changes={self.changes})"

    def __str__(self):
        return "\n".join(sorted([str(line) for line in self.changes]))


class FullyQualifyNames(ast.NodeTransformer):
    """Transform a statement to a one where external names are fully qualified.

    Example:
        Given a statement 'a = join(...)' and an information that join() was
        imported by 'from os.path import join' statement, produce
        'a = os.path.join(...)' statement. The visitor also records which
        substitutions were made.
    """

    def __init__(self, imports: pi.ImportedNames) -> None:
        super(FullyQualifyNames, self).__init__()
        self.external_names: pi.ImportedNames = imports
        self.substitutions: Dict[str, str] = {}
        self.references: Dict[str, str] = {}
        self._current = None

    def visit_Name(self, node):  # pylint: disable=invalid-name, missing-docstring
        self._current = None
        if node.id in self.external_names:
            self._current = self.external_names[node.id].canonical_name
            self.references[self._current] = node.id
            if node.id == self._current:
                return node

            self.substitutions[node.id] = self._current
            return self.external_names[node.id].canonical_ast

        return node

    def visit_Attribute(self, node):  # pylint: disable=invalid-name, missing-docstring
        self.generic_visit(node)
        if self._current:
            prefix = self._current
            self._current = ".".join([prefix, node.attr])
            key = self.references[prefix]
            self.references[self._current] = ".".join([key, node.attr])

        return node


def find_external_name_matches(
    old: ast.AST, new: ast.AST, old_imports: pi.ImportedNames, new_imports: pi.ImportedNames
) -> Optional[ExternalNameUsageChange]:
    """Tests two statements for semantical equality w.r.t. external name usage.

    Example:
        (1) something = os.path.join(...)
        (2) something = path.join(...)

        Provided statement (1) runs in context where 'import os.path' was
        executed and (2) runs in context where 'from os import path' was imported,
        the two statements are semantically identical.

    Args:
        old: AST of the old version of the statement
        new: AST of the new version of the statement
        old_imports: Imported names available for old version of the enclosing functions
        new_imports: Imported names available for new version of the enclosing functions

    Returns:
        If the statements are identical, returns None. If the statements differ
        in something else than external name usage, returns None. Otherwise,
        returns a set of ExternalInStmtChange objects, each
        representing a single external name usage change."""

    if ast.dump(old) == ast.dump(new):
        return None

    fq_old_transformer = FullyQualifyNames(old_imports)
    fq_new_transformer = FullyQualifyNames(new_imports)

    LOGGER.debug("Fully qualifying old statement")
    fq_old = fq_old_transformer.visit(copy.deepcopy(old))
    LOGGER.debug("Fully qualifying new statement")
    fq_new = fq_new_transformer.visit(copy.deepcopy(new))

    changes: Set[SingleExternalNameUsageChange] = set()

    if ast.dump(fq_old) == ast.dump(fq_new):
        LOGGER.debug("Statements are identical after full qualification")
        LOGGER.debug(f"Old statement references: {fq_old_transformer.references}")
        LOGGER.debug(f"New statement references: {fq_new_transformer.references}")
        for original, fqdn in fq_old_transformer.substitutions.items():
            LOGGER.debug(f"'{original}' was fully qualified as '{fqdn}' in old statement")
            if fqdn in fq_new_transformer.references:
                changes.add(
                    SingleExternalNameUsageChange(original, fq_new_transformer.references[fqdn])
                )

        for original, fqdn in fq_new_transformer.substitutions.items():
            LOGGER.debug(f"'{original}' was fully qualified as '{fqdn}' in new statement")
            if fqdn in fq_old_transformer.references:
                changes.add(
                    SingleExternalNameUsageChange(fq_old_transformer.references[fqdn], original)
                )

    return ExternalNameUsageChange(changes) if changes else None


class StatementPyfference:
    """Describes differences between two statements."""

    def __init__(self) -> None:
        # These functions are intentionally not typed
        self.semantically_relevant: Set = set()
        self.semantically_irrelevant: Set = set()

    def add_semantically_irrelevant_change(self, change) -> None:  # pylint: disable=invalid-name
        """Adds semantically irrelevant change."""
        self.semantically_irrelevant.add(change)

    def add_semantically_relevant_change(self, change) -> None:  # pylint: disable=invalid-name
        """Adds semantically relevant change."""
        self.semantically_relevant.add(change)

    def semantically_different(self) -> bool:
        """Returns whether the differences make the statements semantically different.

        Returns:
            If all known changes are semantically irrelevant, return False.
            Otherwise, return True. If there are no known changes, assume the statements
            were semantically different."""
        # Either we know we have some semantically relevant change,
        # or we have NO IDENTIFIED change, therefore we must assume
        # the difference is semantically relevant
        return bool(self.semantically_relevant or not self.semantically_irrelevant)

    def __str__(self):
        lines = []
        for change in self.semantically_relevant:
            lines.append(str(change))

        for change in self.semantically_irrelevant:
            lines.append(str(change))

        return "\n".join(lines)

    def __repr__(self):
        return (
            f"StatementPyfference(relevant={repr(self.semantically_relevant)}, "
            f"irrelevant={repr(self.semantically_irrelevant)})"
        )

    def is_specific(self) -> bool:
        """Returns True if we have at least one identified code change."""
        return bool(self.semantically_relevant or self.semantically_irrelevant)


def pyff_statement(
    old_statement: ast.AST,
    new_statement: ast.AST,
    old_imports: pi.ImportedNames,
    new_imports: pi.ImportedNames,
) -> Optional[StatementPyfference]:
    """Compare two statements.

    Args:
        old_statement: Old version of the statement
        new_statement: New version of the statement
        old_imports: Imported names available for old version of the statement
        new_imports: Imported names available for new version of the statement

    Returns:
        If the statements are identical, returns None. If they differ, a StatementPyfference
        object, describing the differences is returned."""

    if ast.dump(old_statement) == ast.dump(new_statement):
        return None

    pyfference = StatementPyfference()

    change = find_external_name_matches(old_statement, new_statement, old_imports, new_imports)
    if change:
        LOGGER.debug("Statements are semantically identical but differ in imported name references")
        pyfference.add_semantically_irrelevant_change(change)

    return pyfference



"""Entry point for the `pyff` command"""

import sys
import logging
import pathlib
from argparse import ArgumentParser
from typing import Callable

from pyff.modules import pyff_module_code
from pyff.packages import pyff_package
from pyff.kitchensink import highlight, HIGHLIGHTS

LOGGER = logging.getLogger(__name__)


def _pyff_that(function: Callable, what: str) -> None:
    parser = ArgumentParser()
    parser.add_argument("old")
    parser.add_argument("new")

    parser.add_argument("--highlight-names", dest="highlight", choices=HIGHLIGHTS, default="color")
    parser.add_argument("--debug", action="store_true", default=False)

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(
            format="%(levelname)s:%(name)s:%(funcName)s: %(message)s", level=logging.DEBUG
        )

    LOGGER.debug(f"Python Diff: old {what} {args.old} | new {what} {args.new}")
    changes = function(pathlib.Path(args.old), pathlib.Path(args.new))

    if changes is None:
        print(f"Pyff did not detect any significant difference between {args.old} and {args.new}")
        sys.exit(0)

    print(highlight(str(changes), args.highlight))


def pyffmod() -> None:
    """Entry point for the `pyff` command"""

    def compare(old, new):
        """Open two arguments as files and compare them"""
        with open(old, "r") as old_module, open(new, "r") as new_module:
            old_version = old_module.read()
            new_version = new_module.read()

        return pyff_module_code(old_version, new_version)

    _pyff_that(compare, "module")


def pyffpkg() -> None:
    """Entry point for the `pyff-package` command"""
    _pyff_that(pyff_package, "package")


if __name__ == "__main__":
    pyffmod()
